@@ -1,4 +1,4 @@
-# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE
+# BSD 3-Clause License; see https://github.com/scikit-hep/uproot5/blob/main/LICENSE
 
 """
 Uproot: ROOT I/O in pure Python and NumPy.
@@ -18,10 +18,10 @@ Typical entry points for file-reading are
 * :doc:`uproot.reading.open`
 * :doc:`uproot.behaviors.TBranch.iterate`
 * :doc:`uproot.behaviors.TBranch.concatenate`
-* :doc:`uproot.behaviors.TBranch.lazy`
+* :doc:`uproot._dask.dask`
 
 though they would usually be accessed as ``uproot.iterate``,
-``uproot.concatenate``, and ``uproot.lazy``.
+``uproot.concatenate``, and ``uproot.dask``.
 
 The most useful classes are
 
@@ -72,7 +72,6 @@ The submodules of Uproot are:
 isort:skip_file
 """
 
-
 from uproot.version import __version__
 import uproot.const
 import uproot.extras
@@ -90,7 +89,9 @@ from uproot.source.http import HTTPSource
 from uproot.source.http import MultithreadedHTTPSource
 from uproot.source.xrootd import XRootDSource
 from uproot.source.xrootd import MultithreadedXRootDSource
+from uproot.source.s3 import S3Source
 from uproot.source.object import ObjectSource
+from uproot.source.fsspec import FSSpecSource
 from uproot.source.cursor import Cursor
 from uproot.source.futures import TrivialExecutor
 from uproot.source.futures import ThreadPoolExecutor
@@ -122,6 +123,9 @@ from uproot.writing import WritableDirectory
 from uproot.writing import WritableTree
 from uproot.writing import WritableBranch
 from uproot.writing import to_writable
+from uproot.writing import dask_write
+
+from uproot.writing.interpret import as_TGraph
 
 import uproot.models.TObject
 import uproot.models.TString
@@ -145,6 +149,7 @@ import uproot.models.RNTuple
 import uproot.models.TH
 import uproot.models.TGraph
 import uproot.models.TMatrixT
+import uproot.models.TTime
 
 from uproot.models.TTree import num_entries
 
@@ -180,11 +185,11 @@ from uproot.behaviors.TTree import TTree
 from uproot.behaviors.TBranch import TBranch
 from uproot.behaviors.TBranch import iterate
 from uproot.behaviors.TBranch import concatenate
-from uproot.behaviors.TBranch import lazy
-from uproot.behaviors.TBranch import dask
 
 from uproot.behavior import behavior_of
 
 from uproot._util import no_filter
+from uproot._dask import dask, ImplementsFormMapping, ImplementsFormMappingInfo
 
 from uproot.pyroot import from_pyroot
+from uproot.pyroot import to_pyroot
