@@ -20,6 +20,10 @@ import numpy
 
 import uproot
 from uproot._util import no_filter, unset
+from uproot.behaviors.RNTuple import HasFields
+from uproot.behaviors.RNTuple import (
+    _regularize_step_size as _RNTuple_regularize_step_size,
+)
 from uproot.behaviors.TBranch import HasBranches, TBranch, _regularize_step_size
 
 if TYPE_CHECKING:
@@ -87,8 +91,10 @@ def dask(
             of dask arrays and if ``library='ak'`` it returns a single dask-awkward
             array. ``library='pd'`` has not been implemented yet and will raise a
             ``NotImplementedError``.
-        ak_add_doc (bool): If True and ``library="ak"``, add the TBranch ``title``
+        ak_add_doc (bool | dict ): If True and ``library="ak"``, add the TBranch ``title``
             to the Awkward ``__doc__`` parameter of the array.
+            if dict = {key:value} and ``library="ak"``, add the TBranch ``value`` to the
+            Awkward ``key`` parameter of the array.
         custom_classes (None or dict): If a dict, override the classes from
             the :doc:`uproot.reading.ReadOnlyFile` or ``uproot.classes``.
         allow_missing (bool): If True, skip over any files that do not contain
@@ -614,7 +620,13 @@ def _get_dask_array(
                 recursive=recursive,
                 filter_name=filter_name,
                 filter_typename=filter_typename,
-                filter_branch=real_filter_branch,
+                **{
+                    (
+                        "filter_field"
+                        if isinstance(obj, HasFields)
+                        else "filter_branch"
+                    ): real_filter_branch
+                },
                 full_paths=full_paths,
                 ignore_duplicates=True,
             )
@@ -630,9 +642,7 @@ def _get_dask_array(
         assert steps_per_file is not unset  # either assigned or assumed to be 1
         total_files = len(ttrees)
         total_entries = sum(ttree.num_entries for ttree in ttrees)
-        step_size = max(
-            1, int(math.ceil(total_entries / (total_files * steps_per_file)))
-        )
+        step_size = max(1, math.ceil(total_entries / (total_files * steps_per_file)))
 
     if count == 0:
         raise ValueError(
@@ -670,7 +680,7 @@ def _get_dask_array(
         )
         step_sum += int(ttree_step)
 
-    entry_step = int(round(step_sum / len(ttrees)))
+    entry_step = round(step_sum / len(ttrees))
     assert entry_step >= 1
 
     for key in common_keys:
@@ -756,7 +766,11 @@ def _get_dask_array_delay_open(
         recursive=recursive,
         filter_name=filter_name,
         filter_typename=filter_typename,
-        filter_branch=filter_branch,
+        **{
+            (
+                "filter_field" if isinstance(obj, HasFields) else "filter_branch"
+            ): filter_branch
+        },
         full_paths=full_paths,
         ignore_duplicates=True,
     )
@@ -925,6 +939,13 @@ class TrivialFormMappingInfo(ImplementsFormMappingInfo):
             how=tuple,
         )
 
+        if isinstance(tree, HasFields):
+            # Temporary workaround to have basic support for RNTuple
+            # This is needed since currently arrays only has top-level fields
+            # TODO: Ask people how they want this handled since the previous
+            # approach might not make sense for RNTuples
+            keys = [f for f in tree.field_names if f in keys]
+
         awkward = uproot.extras.awkward()
 
         # The subform generated by awkward.to_buffers() has different form keys
@@ -1084,6 +1105,9 @@ class UprootReadMixin:
         keys = self.necessary_columns(report=report, state=state)
         return self.project_keys(keys)
 
+    def project_manually(self: T, columns: frozenset[str]) -> T:
+        return self.project_keys(columns)
+
     def necessary_columns(
         self, *, report: TypeTracerReport, state: dict
     ) -> frozenset[str]:
@@ -1383,11 +1407,29 @@ def _get_ttree_form(
     for key in common_keys:
         branch = ttree[key]
         content_form = branch.interpretation.awkward_form(ttree.file)
-        if ak_add_doc:
-            content_form = content_form.copy(parameters={"__doc__": branch.title})
+        content_parameters = {}
+        if isinstance(ak_add_doc, bool):
+            if ak_add_doc:
+                content_parameters["__doc__"] = branch.title
+        elif isinstance(ak_add_doc, dict):
+            content_parameters.update(
+                {
+                    key: branch.__getattribute__(value)
+                    for key, value in ak_add_doc.items()
+                }
+            )
+        if len(content_parameters.keys()) != 0:
+            content_form = content_form.copy(parameters=content_parameters)
         contents.append(content_form)
 
-    parameters = {"__doc__": ttree.title} if ak_add_doc else None
+    if isinstance(ak_add_doc, bool):
+        parameters = {"__doc__": ttree.title} if ak_add_doc else None
+    elif isinstance(ak_add_doc, dict):
+        parameters = (
+            {"__doc__": ttree.title} if "__doc__" in ak_add_doc.keys() else None
+        )
+    else:
+        parameters = None
 
     return awkward.forms.RecordForm(contents, common_keys, parameters=parameters)
 
@@ -1451,7 +1493,13 @@ def _get_dak_array(
                 recursive=recursive,
                 filter_name=filter_name,
                 filter_typename=filter_typename,
-                filter_branch=real_filter_branch,
+                **{
+                    (
+                        "filter_field"
+                        if isinstance(obj, HasFields)
+                        else "filter_branch"
+                    ): real_filter_branch
+                },
                 full_paths=full_paths,
                 ignore_duplicates=True,
             )
@@ -1467,9 +1515,7 @@ def _get_dak_array(
         assert steps_per_file is not unset  # either assigned or assumed to be 1
         total_files = len(ttrees)
         total_entries = sum(ttree.num_entries for ttree in ttrees)
-        step_size = max(
-            1, int(math.ceil(total_entries / (total_files * steps_per_file)))
-        )
+        step_size = max(1, math.ceil(total_entries / (total_files * steps_per_file)))
 
     if count == 0:
         raise ValueError(
@@ -1496,16 +1542,23 @@ def _get_dak_array(
         entry_start = 0
         entry_stop = ttree.num_entries
 
-        branchid_interpretation = {}
-        for key in common_keys:
-            branch = ttree[key]
-            branchid_interpretation[branch.cache_key] = branch.interpretation
-        ttree_step = _regularize_step_size(
-            ttree, step_size, entry_start, entry_stop, branchid_interpretation
-        )
-        step_sum += int(ttree_step)
+        if isinstance(ttree, HasFields):
+            akform = ttree.to_akform(filter_name=common_keys)
+            ttree_step = _RNTuple_regularize_step_size(
+                ttree, akform, step_size, entry_start, entry_stop
+            )
+            step_sum += int(ttree_step)
+        else:
+            branchid_interpretation = {}
+            for key in common_keys:
+                branch = ttree[key]
+                branchid_interpretation[branch.cache_key] = branch.interpretation
+            ttree_step = _regularize_step_size(
+                ttree, step_size, entry_start, entry_stop, branchid_interpretation
+            )
+            step_sum += int(ttree_step)
 
-    entry_step = int(round(step_sum / len(ttrees)))
+    entry_step = round(step_sum / len(ttrees))
 
     divisions = [0]
     partition_args = []
@@ -1539,9 +1592,12 @@ which has {entry_stop} entries"""
                     divisions.append(divisions[-1] + length)
                     partition_args.append((i, start, stop))
 
-    base_form = _get_ttree_form(
-        awkward, ttrees[0], common_keys, interp_options.get("ak_add_doc")
-    )
+    if isinstance(ttrees[0], HasFields):
+        base_form = ttrees[0].to_akform(filter_name=common_keys)
+    else:
+        base_form = _get_ttree_form(
+            awkward, ttrees[0], common_keys, interp_options.get("ak_add_doc")
+        )
 
     if len(partition_args) == 0:
         divisions.append(0)
@@ -1609,7 +1665,11 @@ def _get_dak_array_delay_open(
             recursive=recursive,
             filter_name=filter_name,
             filter_typename=filter_typename,
-            filter_branch=filter_branch,
+            **{
+                (
+                    "filter_field" if isinstance(obj, HasFields) else "filter_branch"
+                ): filter_branch
+            },
             full_paths=full_paths,
             ignore_duplicates=True,
         )
