@@ -1,10 +1,11 @@
-# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE
+# BSD 3-Clause License; see https://github.com/scikit-hep/uproot5/blob/main/LICENSE
 
 """
 This module defines an interface to compression algorithms used by ROOT, as well
 as functions for compressing and decompressing a :doc:`uproot.source.chunk.Chunk`.
 """
 
+from __future__ import annotations
 
 import struct
 
@@ -86,11 +87,30 @@ class _DecompressZLIB:
     name = "ZLIB"
     _2byte = b"ZL"
     _method = b"\x08"
+    library = "zlib"  # options: "zlib", "isal", "deflate"
 
-    def decompress(self, data, uncompressed_bytes=None):
-        import zlib
+    def decompress(self, data: bytes, uncompressed_bytes=None) -> bytes:
+        if uncompressed_bytes is None:
+            raise ValueError(
+                "zlib decompression requires the number of uncompressed bytes"
+            )
+        if self.library == "zlib":
+            import zlib
 
-        return zlib.decompress(data)
+            return zlib.decompress(data, bufsize=uncompressed_bytes)
+
+        elif self.library == "isal":
+            isal_zlib = uproot.extras.isal().isal_zlib
+            return isal_zlib.decompress(data, bufsize=uncompressed_bytes)
+
+        elif self.library == "deflate":
+            deflate = uproot.extras.deflate()
+            return deflate.zlib_decompress(data, bufsize=uncompressed_bytes)
+
+        else:
+            raise ValueError(
+                f"unrecognized ZLIB.library: {self.library!r}; must be one of ['zlib', 'isal', 'deflate']"
+            )
 
 
 class ZLIB(Compression, _DecompressZLIB):
@@ -101,7 +121,12 @@ class ZLIB(Compression, _DecompressZLIB):
 
     Represents the ZLIB compression algorithm.
 
-    Uproot uses ``zlib`` from the Python standard library.
+    If ``ZLIB.library`` is ``"zlib"`` (default), Uproot uses ``zlib`` from the
+    Python standard library.
+
+    If ``ZLIB.library`` is ``"isal"``, Uproot uses ``isal.isal_zlib``.
+
+    If ``ZLIB.library`` is ``"deflate"``, Uproot uses ``deflate.deflate_zlib``.
     """
 
     def __init__(self, level):
@@ -113,6 +138,10 @@ class ZLIB(Compression, _DecompressZLIB):
         """
         The compression level: 0 is uncompressed, 1 is minimally compressed, and
         9 is maximally compressed.
+
+        This value to adapted to the ISAL compression levels if that library is used.
+        Note: with ISAL 0 is lowest compression, not uncompressed!
+              as such, we don't allow 0 in isal mode for compatibility reasons.
         """
         return self._level
 
@@ -124,10 +153,34 @@ class ZLIB(Compression, _DecompressZLIB):
             raise ValueError("Compression level must be between 0 and 9 (inclusive)")
         self._level = int(value)
 
-    def compress(self, data):
-        import zlib
+    def compress(self, data: bytes) -> bytes:
+        if self.library == "zlib":
+            import zlib
 
-        return zlib.compress(data, self._level)
+            return zlib.compress(data, level=self._level)
+
+        elif self.library == "isal":
+            isal_zlib = uproot.extras.isal().isal_zlib
+            if self._level == 0:
+                raise ValueError(
+                    'ZLIB.library="isal", and therefore requesting no compression '
+                    "implicitly with level 0 is not allowed."
+                )
+            return isal_zlib.compress(data, level=round(self._level / 3))
+
+        elif self.library == "deflate":
+            deflate = uproot.extras.deflate()
+            if self._level == 0:
+                raise ValueError(
+                    'ZLIB.library="deflate", and therefore requesting no compression '
+                    "implicitly with level 0 is not allowed."
+                )
+            return deflate.zlib_compress(data, round(self._level))
+
+        else:
+            raise ValueError(
+                f"unrecognized ZLIB.library: {self.library!r}; must be one of ['zlib', 'isal', 'deflate']"
+            )
 
 
 class _DecompressLZMA:
@@ -135,9 +188,20 @@ class _DecompressLZMA:
     _2byte = b"XZ"
     _method = b"\x00"
 
-    def decompress(self, data, uncompressed_bytes=None):
-        lzma = uproot.extras.lzma()
-        return lzma.decompress(data)
+    def decompress(self, data: bytes, uncompressed_bytes=None) -> bytes:
+        cramjam = uproot.extras.cramjam()
+        lzma = getattr(cramjam, "xz", None) or getattr(
+            getattr(cramjam, "experimental", None), "lzma", None
+        )
+        if lzma is None:
+            import lzma
+
+            return lzma.decompress(data)
+        if uncompressed_bytes is None:
+            raise ValueError(
+                "lzma decompression requires the number of uncompressed bytes"
+            )
+        return lzma.decompress(data, output_len=uncompressed_bytes)
 
 
 class LZMA(Compression, _DecompressLZMA):
@@ -148,9 +212,7 @@ class LZMA(Compression, _DecompressLZMA):
 
     Represents the LZMA compression algorithm.
 
-    Uproot uses ``lzma`` from the Python 3 standard library.
-
-    In Python 2, ``backports.lzma`` must be installed.
+    Uproot uses ``lzma`` from the ``cramjam`` package.
     """
 
     def __init__(self, level):
@@ -173,8 +235,13 @@ class LZMA(Compression, _DecompressLZMA):
             raise ValueError("Compression level must be between 0 and 9 (inclusive)")
         self._level = int(value)
 
-    def compress(self, data):
-        lzma = uproot.extras.lzma()
+    def compress(self, data: bytes) -> bytes:
+        cramjam = uproot.extras.cramjam()
+        lzma = getattr(cramjam, "xz", None) or getattr(
+            getattr(cramjam, "experimental", None), "lzma", None
+        )
+        if lzma is None:
+            import lzma
         return lzma.compress(data, preset=self._level)
 
 
@@ -183,13 +250,13 @@ class _DecompressLZ4:
     _2byte = b"L4"
     _method = b"\x01"
 
-    def decompress(self, data, uncompressed_bytes=None):
-        lz4_block = uproot.extras.lz4_block()
+    def decompress(self, data: bytes, uncompressed_bytes=None) -> bytes:
+        lz4 = uproot.extras.cramjam().lz4
         if uncompressed_bytes is None:
             raise ValueError(
                 "lz4 block decompression requires the number of uncompressed bytes"
             )
-        return lz4_block.decompress(data, uncompressed_size=uncompressed_bytes)
+        return lz4.decompress_block(data, output_len=uncompressed_bytes)
 
 
 class LZ4(Compression, _DecompressLZ4):
@@ -200,7 +267,7 @@ class LZ4(Compression, _DecompressLZ4):
 
     Represents the LZ4 compression algorithm.
 
-    The ``zl4`` and ``xxhash`` libraries must be installed.
+    The ``cramjam`` and ``xxhash`` libraries must be installed.
     """
 
     def __init__(self, level):
@@ -223,9 +290,9 @@ class LZ4(Compression, _DecompressLZ4):
             raise ValueError("Compression level must be between 0 and 12 (inclusive)")
         self._level = int(value)
 
-    def compress(self, data):
-        lz4_block = uproot.extras.lz4_block()
-        return lz4_block.compress(data, compression=self._level, store_size=False)
+    def compress(self, data: bytes) -> bytes:
+        lz4 = uproot.extras.cramjam().lz4
+        return lz4.compress_block(data, compression=self._level, store_size=False)
 
 
 class _DecompressZSTD:
@@ -233,18 +300,13 @@ class _DecompressZSTD:
     _2byte = b"ZS"
     _method = b"\x01"
 
-    def __init__(self):
-        self._decompressor = None
-
-    @property
-    def decompressor(self):
-        if self._decompressor is None:
-            zstandard = uproot.extras.zstandard()
-            self._decompressor = zstandard.ZstdDecompressor()
-        return self._decompressor
-
-    def decompress(self, data, uncompressed_bytes=None):
-        return self.decompressor.decompress(data)
+    def decompress(self, data: bytes, uncompressed_bytes=None) -> bytes:
+        zstd = uproot.extras.cramjam().zstd
+        if uncompressed_bytes is None:
+            raise ValueError(
+                "zstd block decompression requires the number of uncompressed bytes"
+            )
+        return zstd.decompress(data, output_len=uncompressed_bytes)
 
 
 class ZSTD(Compression, _DecompressZSTD):
@@ -255,7 +317,7 @@ class ZSTD(Compression, _DecompressZSTD):
 
     Represents the ZSTD compression algorithm.
 
-    The ``zstandard`` library must be installed.
+    The ``cramjam`` library must be installed.
     """
 
     def __init__(self, level):
@@ -279,15 +341,9 @@ class ZSTD(Compression, _DecompressZSTD):
             raise ValueError("Compression level must be between 0 and 22 (inclusive)")
         self._level = int(value)
 
-    @property
-    def compressor(self):
-        if self._compressor is None:
-            zstandard = uproot.extras.zstandard()
-            self._compressor = zstandard.ZstdCompressor(level=self._level)
-        return self._compressor
-
-    def compress(self, data):
-        return self.compressor.compress(data)
+    def compress(self, data: bytes) -> bytes:
+        zstd = uproot.extras.cramjam().zstd
+        return zstd.compress(data, level=self._level)
 
 
 algorithm_codes = {
@@ -378,10 +434,8 @@ def decompress(
             computed_checksum = xxhash.xxh64(data).intdigest()
             if computed_checksum != expected_checksum:
                 raise ValueError(
-                    """computed checksum {} didn't match expected checksum {}
-in file {}""".format(
-                        computed_checksum, expected_checksum, chunk.source.file_path
-                    )
+                    f"""computed checksum {computed_checksum} didn't match expected checksum {expected_checksum}
+in file {chunk.source.file_path}"""
                 )
 
         elif algo == _decompress_ZSTD._2byte:
@@ -390,19 +444,15 @@ in file {}""".format(
 
         elif algo == b"CS":
             raise ValueError(
-                """unsupported compression algorithm: {} (according to """
-                """ROOT comments, it hasn't been used in 20 years!
-in file {}""".format(
-                    algo, chunk.source.file_path
-                )
+                f"""unsupported compression algorithm: {algo} (according to """
+                f"""ROOT comments, it hasn't been used in 20 years!
+in file {chunk.source.file_path}"""
             )
 
         else:
             raise ValueError(
-                """unrecognized compression algorithm: {}
-in file {}""".format(
-                    algo, chunk.source.file_path
-                )
+                f"""unrecognized compression algorithm: {algo}
+in file {chunk.source.file_path}"""
             )
 
         if block_info is not None:
@@ -416,16 +466,10 @@ in file {}""".format(
 
         if len(uncompressed_bytestring) != block_uncompressed_bytes:
             raise ValueError(
-                """after successfully decompressing {} blocks, a block of """
-                """compressed size {} decompressed to {} bytes, but the """
-                """block header expects {} bytes.
-in file {}""".format(
-                    num_blocks,
-                    block_compressed_bytes,
-                    len(uncompressed_bytestring),
-                    block_uncompressed_bytes,
-                    chunk.source.file_path,
-                )
+                f"""after successfully decompressing {num_blocks} blocks, a block of """
+                f"""compressed size {block_compressed_bytes} decompressed to {len(uncompressed_bytestring)} bytes, but the """
+                f"""block header expects {block_uncompressed_bytes} bytes.
+in file {chunk.source.file_path}"""
             )
 
         uncompressed_array = numpy.frombuffer(
@@ -462,11 +506,11 @@ in file {}""".format(
     return uproot.source.chunk.Chunk.wrap(chunk.source, output)
 
 
-def hook_before_block(**kwargs):  # noqa: D103
+def hook_before_block(**kwargs):
     pass
 
 
-def hook_after_block(**kwargs):  # noqa: D103
+def hook_after_block(**kwargs):
     pass
 
 
@@ -477,7 +521,7 @@ _3BYTE_MAX = 2**24 - 1
 _4byte = struct.Struct("<I")  # compressed sizes are 3-byte little endian!
 
 
-def compress(data, compression):
+def compress(data: bytes, compression: Compression) -> bytes:
     """
     Args:
         data (bytes, memoryview, or NumPy array): Data to compress.
