@@ -1,9 +1,9 @@
-# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE
+# BSD 3-Clause License; see https://github.com/scikit-hep/uproot5/blob/main/LICENSE
 
 """
 This module defines a versionless model for ``THashList``.
 """
-
+from __future__ import annotations
 
 import uproot
 
@@ -17,12 +17,12 @@ class Model_THashList(uproot.model.Model):
         pass
 
     def read_members(self, chunk, cursor, context, file):
+        if uproot._awkwardforth.get_forth_obj(context) is not None:
+            raise uproot.interpretation.objects.CannotBeForth()
         if self.is_memberwise:
             raise NotImplementedError(
-                """memberwise serialization of {}
-in file {}""".format(
-                    type(self).__name__, self.file.file_path
-                )
+                f"""memberwise serialization of {type(self).__name__}
+in file {self.file.file_path}"""
             )
         self._bases.append(
             uproot.models.TList.Model_TList.read(
@@ -41,12 +41,7 @@ in file {}""".format(
             version = ""
         else:
             version = f" (version {self.class_version})"
-        return "<{}{} of {} items at 0x{:012x}>".format(
-            self.classname,
-            version,
-            len(self),
-            id(self),
-        )
+        return f"<{self.classname}{version} of {len(self)} items at 0x{id(self):012x}>"
 
     def __getitem__(self, where):
         return self._bases[0][where]
@@ -54,5 +49,12 @@ in file {}""".format(
     def __len__(self):
         return len(self._bases[0])
 
+    writable = True
+
+    def _serialize(self, out, header, name, tobject_flags):
+        assert len(self._bases) == 1, "Fatal error on THashList serialization."
+        for x in self._bases:
+            x._serialize(out, True, None, tobject_flags)
+
 
 uproot.classes["THashList"] = Model_THashList
