@@ -1581,8 +1581,8 @@ class Directory(CascadeNode):
         if replaces is None:
             next_key = key.copy_to(self._data.next_location)
             if self._data.num_bytes + next_key.num_bytes > self._data.allocation:
-                requested_num_bytes = int(
-                    math.ceil(1.5 * (self._data.allocation + next_key.num_bytes + 8))
+                requested_num_bytes = math.ceil(
+                    1.5 * (self._data.allocation + next_key.num_bytes + 8)
                 )
                 self._reallocate_data(requested_num_bytes)
                 next_key = key.copy_to(self._data.next_location)
@@ -1597,8 +1597,8 @@ class Directory(CascadeNode):
                 self._data.num_bytes + new_key.num_bytes - original_key.num_bytes
                 > self._data.allocation
             ):
-                requested_num_bytes = int(
-                    math.ceil(1.5 * (self._data.allocation + new_key.num_bytes + 8))
+                requested_num_bytes = math.ceil(
+                    1.5 * (self._data.allocation + new_key.num_bytes + 8)
                 )
                 self._reallocate_data(requested_num_bytes)
                 original_key = self._data.get_key(replaces.name.string, replaces.cycle)
@@ -1685,7 +1685,7 @@ class Directory(CascadeNode):
         next_key = subdirectory_key.copy_to(self._data.next_location)
         if self._data.num_bytes + next_key.num_bytes > self._data.allocation:
             self._reallocate_data(
-                int(math.ceil(1.5 * (self._data.allocation + next_key.num_bytes + 8)))
+                math.ceil(1.5 * (self._data.allocation + next_key.num_bytes + 8))
             )
             next_key = subdirectory_key.copy_to(self._data.next_location)
         next_key._location = self._data.next_location
@@ -1727,63 +1727,32 @@ class Directory(CascadeNode):
         tree.write_anew(sink)
         return tree
 
-    def add_rntuple(self, sink, name, title, akform):
+    def add_rntuple(self, sink, name, title, akform, description=""):
         import uproot.writing._cascadentuple
 
         anchor = uproot.writing._cascadentuple.NTuple_Anchor(
-            None, 0, 0, 48, None, None, None, None, None, None, 0
+            None,
+            *uproot.const.rntuple_version_for_writing,
+            None,
+            None,
+            None,
+            None,
+            None,
+            None,
+            0,  # TODO: Fix this
         )
 
-        header = uproot.writing._cascadentuple.NTuple_Header(None, name, "", akform)
-
-        footer = uproot.writing._cascadentuple.NTuple_Footer(
-            None, 0, header._crc32, akform
+        header = uproot.writing._cascadentuple.NTuple_Header(
+            None, name, description, akform
         )
 
-        # the empty page list is hard-coded bytes which represents:
-        # 0                   1                   2                   3
-        # 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
-        # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
-        # |        Envelope Version       |        Minimum Version        |
-        # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
-        # |                             Size                            |T|
-        # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
-        # |           Number of Items (for list frames)           |Reserv.|
-        # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
-        # |                         FRAME PAYLOAD                         |
-        # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
-        # |                             CRC32                             |
-        # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
-        #
-        # - Envelope Version = 1 (0x0100)
-        # - Minimum Version = 1  (0x0100)
-        # - Size = -8 (0xf8ffffff) [value is negative because this is a list]
-        # - Number of Items = 0 (0x00000000) [empty list]
-        # - FRAME PAYLOAD = empty [because number of items is 0]
-        # - CRC32 = 2678769841
-        # manually calculate CRC32:
-
-        # In [1]: zlib.crc32(b'\x01\x00\x01\x00\xf8\xff\xff\xff\00\00\00\00')
-        # Out[1]: 2678769841
-        # In [2]: np.array([177, 200, 170, 159], dtype=np.uint8).view("uint32")
-        # Out[2]: array([2678769841], dtype=uint32)
-
-        empty_page_list_bytes = numpy.array(
-            [1, 0, 1, 0, 248, 255, 255, 255, 0, 0, 0, 0, 177, 200, 170, 159],
-            dtype=numpy.uint8,
-        )
-        offset = self._freesegments.allocate(16)
-        footer.cluster_group_record_frames[0].page_list_envlink.locator = (
-            uproot.writing._cascadentuple.NTuple_Locator(16, offset)
-        )
+        footer = uproot.writing._cascadentuple.NTuple_Footer(None, header._checksum)
 
         ntuple = uproot.writing._cascadentuple.NTuple(
-            self, name, title, akform, self._freesegments, header, footer, [], anchor
+            self, akform, self._freesegments, header, footer, [], anchor
         )
 
-        sink.write(offset, empty_page_list_bytes)
         ntuple.write(sink)
-        sink.flush()
         return ntuple
 
 
