@@ -90,7 +90,7 @@ def _cpp_typename(akform, subcall=False):
         typename = _ak_primitive_to_typename_dict[ak_primitive]
         for arr_size in inner_shape[::-1]:
             typename = f"std::array<{typename},{arr_size}>"
-    elif isinstance(akform, awkward.forms.ListOffsetForm):
+    elif isinstance(akform, (awkward.forms.ListOffsetForm, awkward.forms.ListForm)):
         content_typename = _cpp_typename(akform.content, subcall=True)
         typename = f"std::vector<{content_typename}>"
         override_typename = akform.parameters.get("__array__", "")
@@ -192,19 +192,6 @@ def _serialize_envelope_header(type, length):
     return _rntuple_env_header_format.pack(data)
 
 
-def _serialize_rntuple_page_innerlist(items):  # TODO: check
-    n_items = len(items)
-    payload_bytes = b"".join([x.serialize() for x in items])
-    offset = (0).to_bytes(8, "little")
-    compression_setting = (0).to_bytes(4, "little")
-    payload_bytes = b"".join([payload_bytes, offset, compression_setting])
-    size = 4 + 4 + len(payload_bytes)
-    size_bytes = struct.Struct("<i").pack(-size)  # negative size means list
-    # n.b last byte of `n_item bytes` is reserved as of Sep 2022
-    raw_bytes = b"".join([size_bytes, n_items.to_bytes(4, "little"), payload_bytes])
-    return raw_bytes
-
-
 # https://github.com/root-project/root/blob/master/tree/ntuple/v7/doc/specifications.md#field-description
 class NTuple_Field_Description:
     def __init__(
@@ -397,6 +384,37 @@ class NTuple_Header(CascadeLeaf):
                 add_field=field_role == uproot.const.RNTupleFieldRole.COLLECTION,
                 field_name="_0",
             )
+        elif isinstance(akform, awkward.forms.ListForm):
+            type_name = _cpp_typename(akform)
+            field_role = uproot.const.RNTupleFieldRole.COLLECTION
+            if akform.parameters.get("__array__", "") == "string":
+                type_name = "std::string"
+                field_role = uproot.const.RNTupleFieldRole.LEAF
+            field = NTuple_Field_Description(
+                0,
+                0,
+                parent_fid,
+                field_role,
+                0,
+                field_name,
+                type_name,
+                "",
+                "",
+            )
+            self._field_records.append(field)
+            # They are always converted to ListOffsetArrays with Int64 offsets
+            type_num = _ak_primitive_to_num_dict["i64"]
+            type_size = uproot.const.rntuple_col_num_to_size_dict[type_num]
+            col = NTuple_Column_Description(type_num, type_size, field_id, 0, 0)
+            self._column_records.append(col)
+            self._column_keys.append(f"node{self._ak_node_count}-startstop")
+            # content data
+            self._build_field_col_records(
+                akform.content,
+                parent_fid=field_id,
+                add_field=field_role == uproot.const.RNTupleFieldRole.COLLECTION,
+                field_name="_0",
+            )
         elif isinstance(akform, awkward.forms.RecordForm):
             type_name = _cpp_typename(akform)
             field = NTuple_Field_Description(
@@ -634,13 +652,10 @@ class NTuple_EnvLink:
 
 
 class NTuple_PageListEnvelope:
-    def __init__(
-        self, header_checksum, cluster_summaries, page_data, compression_settings=0
-    ):
+    def __init__(self, header_checksum, cluster_summaries, page_data):
         self.header_checksum = header_checksum
         self.cluster_summaries = cluster_summaries
         self.page_data = page_data
-        self.compression_settings = compression_settings
         self._checksum = None
         assert len(cluster_summaries) == len(page_data)
 
@@ -661,7 +676,7 @@ class NTuple_PageListEnvelope:
                                         col[0][2]
                                     ),
                                     _rntuple_column_compression_settings_format.pack(
-                                        self.compression_settings
+                                        col[0][3]
                                     ),
                                 ]
                             ),
@@ -735,20 +750,6 @@ class NTuple_ClusterSummary:
         return f"{type(self).__name__}({self.num_first_entry}, {self.num_entries}, {self.flags})"
 
 
-class NTuple_InnerListLocator:
-    def __init__(self, page_descs):
-        self.page_descs = page_descs
-
-    def serialize(self):
-        # from RNTuple spec:
-        # to save space, the page descriptions (inner items) are not in a record frame.
-        raw_bytes = _serialize_rntuple_page_innerlist(self.page_descs)
-        return raw_bytes
-
-    def __repr__(self):
-        return f"{type(self).__name__}({self.page_descs})"
-
-
 class NTuple_PageDescription:
     def __init__(self, num_entries, locator):
         assert num_entries < (1 << 32)
@@ -951,6 +952,24 @@ class NTuple(CascadeNode):
                 switches["index"] = indices
                 switches["tag"] = tags + 1
                 col_data = switches
+            elif "startstop" in key:
+                # ListArrays need to be converted to ListOffsetArrays
+                barekey = key.split("-")[0]
+                starts = awkward.index.Index(data_buffers[f"{barekey}-starts"])
+                stops = awkward.index.Index(data_buffers[f"{barekey}-stops"])
+                next_barekey = f"node{int(barekey[4:])+1}"
+                content = awkward.contents.NumpyArray(
+                    data_buffers[f"{next_barekey}-data"]
+                )
+                tmp_buffers = awkward.to_buffers(
+                    awkward.contents.ListArray(
+                        starts, stops, content
+                    ).to_ListOffsetArray64()
+                )[2]
+                data_buffers[f"{next_barekey}-data"] = tmp_buffers["node1-data"]
+                col_data = tmp_buffers["node0-offsets"][1:]
+                # no longer need the temporary data
+                del starts, stops, content, tmp_buffers
             else:
                 col_data = data_buffers[key]
             if "offsets" in key:
@@ -964,12 +983,16 @@ class NTuple(CascadeNode):
             raw_data = col_data.reshape(-1).view("uint8")
             if col_data.dtype == numpy.dtype("bool"):
                 raw_data = numpy.packbits(raw_data, bitorder="little")
-            page_key = self.add_rblob(sink, raw_data, len(raw_data), big=False)
+            uncompressed_bytes = len(raw_data)
+            # Need better logic to specify per-column/field compression
+            compression = self._directory.freesegments.fileheader.compression
+            raw_data = uproot.compression.compress(raw_data, compression)
+            page_key = self.add_rblob(sink, raw_data, uncompressed_bytes)
             page_locator = NTuple_Locator(
                 len(raw_data), page_key.location + page_key.allocation
             )
             cluster_page_data.append(
-                [(page_locator, col_len, self._column_counts[idx])]
+                [(page_locator, col_len, self._column_counts[idx], compression.code)]
             )
             self._column_counts[idx] += col_len
         page_data = [
@@ -989,7 +1012,7 @@ class NTuple(CascadeNode):
         )
         pagelistenv_rawdata = pagelistenv.serialize()
         pagelistenv_key = self.add_rblob(
-            sink, pagelistenv_rawdata, len(pagelistenv_rawdata), big=False
+            sink, pagelistenv_rawdata, len(pagelistenv_rawdata)
         )
         pagelistenv_locator = NTuple_Locator(
             len(pagelistenv_rawdata),
@@ -1012,12 +1035,7 @@ class NTuple(CascadeNode):
             old_footer_key.location, old_footer_key.location + old_footer_key.allocation
         )
         footer_raw_data = self._footer.serialize()
-        self._footer_key = self.add_rblob(
-            sink,
-            footer_raw_data,
-            len(footer_raw_data),
-            big=False,
-        )
+        self._footer_key = self.add_rblob(sink, footer_raw_data, len(footer_raw_data))
 
         # 4. Update anchor's foot metadata values in-place
 
@@ -1038,33 +1056,21 @@ class NTuple(CascadeNode):
         sink,
         raw_data,
         uncompressed_bytes,
-        big=None,
     ):
-        strings_size = 8
+        strings_size = 8  # TODO: What is this?
 
-        location = None
-        if not big:
-            requested_bytes = (
-                uproot.reading._key_format_small.size + strings_size + len(raw_data)
-            )
-            location = self._freesegments.allocate(requested_bytes, dry_run=True)
-            if location < uproot.const.kStartBigFile:
-                self._freesegments.allocate(requested_bytes, dry_run=False)
-            else:
-                location = None
-
-        if location is None:
-            requested_bytes = (
-                uproot.reading._key_format_big.size + strings_size + len(raw_data)
-            )
-            location = self._freesegments.allocate(requested_bytes, dry_run=False)
+        # Always use big files
+        requested_bytes = (
+            uproot.reading._key_format_big.size + strings_size + len(raw_data)
+        )
+        location = self._freesegments.allocate(requested_bytes, dry_run=False)
 
         key = RBlob_Key(
             location,
             uncompressed_bytes,
             len(raw_data),
             created_on=datetime.datetime.now(),
-            big=big,
+            big=True,
         )
 
         key.write(sink)
@@ -1076,12 +1082,7 @@ class NTuple(CascadeNode):
     def write(self, sink):
         #### Header ##############################
         header_raw_data = self._header.serialize()
-        self._header_key = self.add_rblob(
-            sink,
-            header_raw_data,
-            len(header_raw_data),
-            big=False,
-        )
+        self._header_key = self.add_rblob(sink, header_raw_data, len(header_raw_data))
         self._anchor.seek_header = (
             self._header_key.location + self._header_key.allocation
         )
@@ -1091,12 +1092,7 @@ class NTuple(CascadeNode):
 
         #### Footer ##############################
         footer_raw_data = self._footer.serialize()
-        self._footer_key = self.add_rblob(
-            sink,
-            footer_raw_data,
-            len(footer_raw_data),
-            big=False,
-        )
+        self._footer_key = self.add_rblob(sink, footer_raw_data, len(footer_raw_data))
         self._anchor.seek_footer = (
             self._footer_key.location + self._footer_key.allocation
         )
