@@ -1,4 +1,4 @@
-# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE
+# BSD 3-Clause License; see https://github.com/scikit-hep/uproot5/blob/main/LICENSE
 
 """
 This module defines a series of rules for identifying Python objects that can be written
@@ -15,13 +15,14 @@ but not adding it to any :doc:`uproot.writing.writable.WritableDirectory`.
 The (many) other functions in this module construct writable :doc:`uproot.model.Model`
 objects from Python builtins and other writable models.
 """
-
+from __future__ import annotations
 
 from collections.abc import Mapping
 
 import numpy
 
 import uproot.compression
+import uproot.extras
 import uproot.pyroot
 import uproot.writing._cascadetree
 
@@ -52,7 +53,9 @@ def add_to_directory(obj, name, directory, streamers):
     if uproot._util.from_module(obj, "pandas"):
         import pandas
 
-        if isinstance(obj, pandas.DataFrame) and obj.index.is_numeric():
+        if isinstance(
+            obj, pandas.DataFrame
+        ) and uproot._util.pandas_has_attr_is_numeric(pandas)(obj.index):
             obj = uproot.writing._cascadetree.dataframe_to_dict(obj)
 
     if uproot._util.from_module(obj, "awkward"):
@@ -64,7 +67,7 @@ def add_to_directory(obj, name, directory, streamers):
     if isinstance(obj, numpy.ndarray) and obj.dtype.fields is not None:
         obj = uproot.writing._cascadetree.recarray_to_dict(obj)
 
-    if isinstance(obj, Mapping) and all(uproot._util.isstr(x) for x in obj):
+    if isinstance(obj, Mapping) and all(isinstance(x, str) for x in obj):
         data = {}
         metadata = {}
 
@@ -73,7 +76,7 @@ def add_to_directory(obj, name, directory, streamers):
                 import pandas
 
                 if isinstance(branch_array, pandas.DataFrame):
-                    branch_array = uproot.writing._cascadetree.dataframe_to_dict(
+                    branch_array = uproot.writing._cascadetree.dataframe_to_dict(  # noqa: PLW2901 (overwriting branch_array)
                         branch_array
                     )
 
@@ -81,21 +84,24 @@ def add_to_directory(obj, name, directory, streamers):
                 isinstance(branch_array, numpy.ndarray)
                 and branch_array.dtype.fields is not None
             ):
-                branch_array = uproot.writing._cascadetree.recarray_to_dict(
+                branch_array = uproot.writing._cascadetree.recarray_to_dict(  # noqa: PLW2901 (overwriting branch_array)
                     branch_array
                 )
 
             if isinstance(branch_array, Mapping) and all(
-                uproot._util.isstr(x) for x in branch_array
+                isinstance(x, str) for x in branch_array
             ):
-                okay = True
                 datum = {}
                 metadatum = {}
                 for kk, vv in branch_array.items():
                     try:
-                        vv = uproot._util.ensure_numpy(vv)
+                        vv = (  # noqa: PLW2901 (overwriting vv)
+                            uproot._util.ensure_numpy(vv)
+                        )
                     except TypeError:
-                        okay = False
+                        raise TypeError(
+                            f"unrecognizable array type {type(branch_array)} associated with {branch_name!r}"
+                        ) from None
                     datum[kk] = vv
                     branch_dtype = vv.dtype
                     branch_shape = vv.shape[1:]
@@ -103,9 +109,6 @@ def add_to_directory(obj, name, directory, streamers):
                         branch_dtype = numpy.dtype((branch_dtype, branch_shape))
                     metadatum[kk] = branch_dtype
 
-                if not okay:
-                    break
-
                 data[branch_name] = datum
                 metadata[branch_name] = metadatum
 
@@ -116,16 +119,19 @@ def add_to_directory(obj, name, directory, streamers):
 
                 else:
                     try:
-                        branch_array = uproot._util.ensure_numpy(branch_array)
+                        branch_array = uproot._util.ensure_numpy(  # noqa: PLW2901 (overwriting branch_array)
+                            branch_array
+                        )
                     except TypeError:
+                        awkward = uproot.extras.awkward()
                         try:
-                            import awkward
-                        except ImportError:
-                            break
-                        try:
-                            branch_array = awkward.from_iter(branch_array)
+                            branch_array = awkward.from_iter(  # noqa: PLW2901 (overwriting branch_array)
+                                branch_array
+                            )
                         except Exception:
-                            break
+                            raise TypeError(
+                                f"unrecognizable array type {type(branch_array)} associated with {branch_name!r}"
+                            ) from None
                         else:
                             data[branch_name] = branch_array
                             metadata[branch_name] = awkward.type(branch_array)
@@ -201,7 +207,7 @@ def to_writable(obj):
         and all(len(x.shape) == 1 for x in obj[1])
         and len(obj[0].shape) == len(obj[1])
     ):
-        obj = (obj[0],) + tuple(obj[1]) + tuple(obj[2:])
+        obj = (obj[0], *tuple(obj[1]), *tuple(obj[2:]))
 
     # This is the big if-elif-else chain of rules
     if isinstance(obj, uproot.model.Model):
@@ -214,12 +220,10 @@ def to_writable(obj):
             return uproot.pyroot._PyROOTWritable(obj)
         else:
             raise TypeError(
-                "only instances of TObject can be written to files, not {}".format(
-                    type(obj).__name__
-                )
+                f"only instances of TObject can be written to files, not {type(obj).__name__}"
             )
 
-    elif uproot._util.isstr(obj):
+    elif isinstance(obj, str):
         return to_TObjString(obj)
 
     elif (
@@ -239,11 +243,7 @@ def to_writable(obj):
         ):
             import boost_histogram
 
-        if obj.kind == "MEAN":
-            raise NotImplementedError(
-                "PlottableHistogram with kind='MEAN' (i.e. profile plots) not supported yet"
-            )
-        elif obj.kind != "COUNT":
+        if obj.kind != "COUNT" and obj.kind != "MEAN":
             raise ValueError(
                 "PlottableHistogram can only be converted to ROOT TH* if kind='COUNT' or 'MEAN'"
             )
@@ -261,14 +261,18 @@ def to_writable(obj):
         try:
             # using flow=True if supported
             data = obj.values(flow=True)
-            fSumw2 = obj.variances(flow=True)
+            fSumw2 = (
+                obj.variances(flow=True)
+                if boost_histogram is None
+                or obj.storage_type == boost_histogram.storage.Weight
+                else None
+            )
 
             # and flow=True is different from flow=False (obj actually has flow bins)
-            if (
-                data.shape == obj.values(flow=False).shape
-                and fSumw2.shape == obj.variances(flow=True).shape
-            ):
-                raise TypeError
+            data_noflow = obj.values(flow=False)
+            for flow, noflow in zip(data.shape, data_noflow.shape):
+                if flow != noflow + 2:
+                    raise TypeError
 
         except TypeError:
             # flow=True is not supported, fallback to allocate-and-fill
@@ -286,19 +290,24 @@ def to_writable(obj):
                 data = numpy.zeros((s[0] + 2, s[1] + 2, s[2] + 2), dtype=d)
                 data[1:-1, 1:-1, 1:-1] = tmp
 
-            tmp = obj.variances()
-            s = tmp.shape
-            if tmp is None:
-                fSumw2 = None
-            elif ndim == 1:
-                fSumw2 = numpy.zeros(s[0] + 2, dtype=">f8")
-                fSumw2[1:-1] = tmp
-            elif ndim == 2:
-                fSumw2 = numpy.zeros((s[0] + 2, s[1] + 2), dtype=">f8")
-                fSumw2[1:-1, 1:-1] = tmp
-            elif ndim == 3:
-                fSumw2 = numpy.zeros((s[0] + 2, s[1] + 2, s[2] + 2), dtype=">f8")
-                fSumw2[1:-1, 1:-1, 1:-1] = tmp
+            tmp = (
+                obj.variances()
+                if boost_histogram is None
+                or obj.storage_type == boost_histogram.storage.Weight
+                else None
+            )
+            fSumw2 = None
+            if tmp is not None:
+                s = tmp.shape
+                if ndim == 1:
+                    fSumw2 = numpy.zeros(s[0] + 2, dtype=">f8")
+                    fSumw2[1:-1] = tmp
+                elif ndim == 2:
+                    fSumw2 = numpy.zeros((s[0] + 2, s[1] + 2), dtype=">f8")
+                    fSumw2[1:-1, 1:-1] = tmp
+                elif ndim == 3:
+                    fSumw2 = numpy.zeros((s[0] + 2, s[1] + 2, s[2] + 2), dtype=">f8")
+                    fSumw2[1:-1, 1:-1, 1:-1] = tmp
 
         else:
             # continuing to use flow=True, because it is supported
@@ -321,113 +330,164 @@ def to_writable(obj):
         # convert all axes in one list comprehension
         axes = [
             to_TAxis(
-                fName=getattr(axis, "name", default_name),
+                fName=default_name,
                 fTitle=getattr(axis, "label", getattr(obj, "name", "")),
                 fNbins=len(axis),
                 fXmin=axis.edges[0],
                 fXmax=axis.edges[-1],
                 fXbins=_fXbins_maybe_regular(axis, boost_histogram),
+                fLabels=_fLabels_maybe_categorical(axis, boost_histogram),
             )
             for axis, default_name in zip(obj.axes, ["xaxis", "yaxis", "zaxis"])
         ]
 
         # make TH1, TH2, TH3 types independently
         if len(axes) == 1:
-            fTsumw, fTsumw2, fTsumwx, fTsumwx2 = _root_stats_1d(
-                obj.values(flow=False), obj.axes[0].edges
-            )
-            return to_TH1x(
-                fName=None,
-                fTitle=title,
-                data=data,
-                fEntries=fEntries,
-                fTsumw=fTsumw,
-                fTsumw2=fTsumw2,
-                fTsumwx=fTsumwx,
-                fTsumwx2=fTsumwx2,
-                fSumw2=fSumw2,
-                fXaxis=axes[0],
-            )
+            if obj.kind == "MEAN":
+                if hasattr(obj, "storage_type"):
+                    if "fSumw2" in obj.metadata.keys():
+                        fSumw2 = obj.metadata["fSumw2"]
+                    else:
+                        raise ValueError(f"fSumw2 has not been set for {obj}")
+                    return to_TProfile(
+                        fName=None,
+                        fTitle=title,
+                        data=obj.values(flow=True),
+                        fEntries=obj.size + 1,
+                        fTsumw=obj.sum()["sum_of_weights"],
+                        fTsumw2=obj.sum()["sum_of_weights_squared"],
+                        fTsumwx=0,
+                        fTsumwx2=0,
+                        fTsumwy=0,
+                        fTsumwy2=0,
+                        fSumw2=fSumw2,
+                        fBinEntries=obj.counts(flow=True),
+                        fBinSumw2=numpy.asarray([], numpy.float64),
+                        fXaxis=axes[0],
+                    )
+                else:
+                    return to_TProfile(
+                        fName=None,
+                        fTitle=title,
+                        data=obj._bases[0]._bases[-1],
+                        fEntries=obj.member("fEntries"),
+                        fTsumw=obj.member("fTsumw"),
+                        fTsumw2=obj.member("fTsumw2"),
+                        fTsumwx=obj.member("fTsumwx"),
+                        fTsumwx2=obj.member("fTsumwx2"),
+                        fTsumwy=obj.member("fTsumwy"),
+                        fTsumwy2=obj.member("fTsumwy2"),
+                        fSumw2=obj.member("fSumw2"),
+                        fBinEntries=obj.member("fBinEntries"),
+                        fBinSumw2=obj.member("fBinSumw2"),
+                        fXaxis=axes[0],
+                    )
+            else:
+                fTsumw, fTsumw2, fTsumwx, fTsumwx2 = _root_stats_1d(
+                    obj.values(flow=False), obj.axes[0].edges
+                )
+                return to_TH1x(
+                    fName=None,
+                    fTitle=title,
+                    data=data,
+                    fEntries=fEntries,
+                    fTsumw=fTsumw,
+                    fTsumw2=fTsumw2,
+                    fTsumwx=fTsumwx,
+                    fTsumwx2=fTsumwx2,
+                    fSumw2=fSumw2,
+                    fXaxis=axes[0],
+                )
 
         elif len(axes) == 2:
-            (
-                fTsumw,
-                fTsumw2,
-                fTsumwx,
-                fTsumwx2,
-                fTsumwy,
-                fTsumwy2,
-                fTsumwxy,
-            ) = _root_stats_2d(
-                obj.values(flow=False), obj.axes[0].edges, obj.axes[1].edges
-            )
-            return to_TH2x(
-                fName=None,
-                fTitle=title,
-                data=data,
-                fEntries=fEntries,
-                fTsumw=fTsumw,
-                fTsumw2=fTsumw2,
-                fTsumwx=fTsumwx,
-                fTsumwx2=fTsumwx2,
-                fTsumwy=fTsumwy,
-                fTsumwy2=fTsumwy2,
-                fTsumwxy=fTsumwxy,
-                fSumw2=fSumw2,
-                fXaxis=axes[0],
-                fYaxis=axes[1],
-            )
+            if obj.kind == "MEAN":
+                raise NotImplementedError(
+                    "2D PlottableHistogram with kind='MEAN' (i.e. 2D profile plots) not supported yet"
+                )
+            else:
+                (
+                    fTsumw,
+                    fTsumw2,
+                    fTsumwx,
+                    fTsumwx2,
+                    fTsumwy,
+                    fTsumwy2,
+                    fTsumwxy,
+                ) = _root_stats_2d(
+                    obj.values(flow=False), obj.axes[0].edges, obj.axes[1].edges
+                )
+                return to_TH2x(
+                    fName=None,
+                    fTitle=title,
+                    data=data,
+                    fEntries=fEntries,
+                    fTsumw=fTsumw,
+                    fTsumw2=fTsumw2,
+                    fTsumwx=fTsumwx,
+                    fTsumwx2=fTsumwx2,
+                    fTsumwy=fTsumwy,
+                    fTsumwy2=fTsumwy2,
+                    fTsumwxy=fTsumwxy,
+                    fSumw2=fSumw2,
+                    fXaxis=axes[0],
+                    fYaxis=axes[1],
+                )
 
         elif len(axes) == 3:
-            (
-                fTsumw,
-                fTsumw2,
-                fTsumwx,
-                fTsumwx2,
-                fTsumwy,
-                fTsumwy2,
-                fTsumwxy,
-                fTsumwz,
-                fTsumwz2,
-                fTsumwxz,
-                fTsumwyz,
-            ) = _root_stats_3d(
-                obj.values(flow=False),
-                obj.axes[0].edges,
-                obj.axes[1].edges,
-                obj.axes[2].edges,
-            )
-            return to_TH3x(
-                fName=None,
-                fTitle=title,
-                data=data,
-                fEntries=fEntries,
-                fTsumw=fTsumw,
-                fTsumw2=fTsumw2,
-                fTsumwx=fTsumwx,
-                fTsumwx2=fTsumwx2,
-                fTsumwy=fTsumwy,
-                fTsumwy2=fTsumwy2,
-                fTsumwxy=fTsumwxy,
-                fTsumwz=fTsumwz,
-                fTsumwz2=fTsumwz2,
-                fTsumwxz=fTsumwxz,
-                fTsumwyz=fTsumwyz,
-                fSumw2=fSumw2,
-                fXaxis=axes[0],
-                fYaxis=axes[1],
-                fZaxis=axes[2],
-            )
+            if obj.kind == "MEAN":
+                raise NotImplementedError(
+                    "3D PlottableHistogram with kind='MEAN' (i.e. 3D profile plots) not supported yet"
+                )
+            else:
+                (
+                    fTsumw,
+                    fTsumw2,
+                    fTsumwx,
+                    fTsumwx2,
+                    fTsumwy,
+                    fTsumwy2,
+                    fTsumwxy,
+                    fTsumwz,
+                    fTsumwz2,
+                    fTsumwxz,
+                    fTsumwyz,
+                ) = _root_stats_3d(
+                    obj.values(flow=False),
+                    obj.axes[0].edges,
+                    obj.axes[1].edges,
+                    obj.axes[2].edges,
+                )
+                return to_TH3x(
+                    fName=None,
+                    fTitle=title,
+                    data=data,
+                    fEntries=fEntries,
+                    fTsumw=fTsumw,
+                    fTsumw2=fTsumw2,
+                    fTsumwx=fTsumwx,
+                    fTsumwx2=fTsumwx2,
+                    fTsumwy=fTsumwy,
+                    fTsumwy2=fTsumwy2,
+                    fTsumwxy=fTsumwxy,
+                    fTsumwz=fTsumwz,
+                    fTsumwz2=fTsumwz2,
+                    fTsumwxz=fTsumwxz,
+                    fTsumwyz=fTsumwyz,
+                    fSumw2=fSumw2,
+                    fXaxis=axes[0],
+                    fYaxis=axes[1],
+                    fZaxis=axes[2],
+                )
 
     elif (
         isinstance(obj, (tuple, list))
         and 2 <= len(obj) <= 5  # might have a histogram title as the last item
         and all(isinstance(x, numpy.ndarray) for x in obj[:-1])
-        and (isinstance(obj[-1], numpy.ndarray) or uproot._util.isstr(obj[-1]))
+        and isinstance(obj[-1], (numpy.ndarray, str))
         and len(obj[0].shape) == sum(int(isinstance(x, numpy.ndarray)) for x in obj[1:])
         and all(len(x.shape) == 1 for x in obj[1:] if isinstance(x, numpy.ndarray))
     ):
-        if uproot._util.isstr(obj[-1]):
+        if isinstance(obj[-1], str):
             obj, title = obj[:-1], obj[-1]
         else:
             title = ""
@@ -660,6 +720,34 @@ def _fXbins_maybe_regular(axis, boost_histogram):
             return axis.edges
 
 
+def _fLabels_maybe_categorical(axis, boost_histogram):
+    if boost_histogram is None:
+        return None
+
+    if not isinstance(
+        axis, (boost_histogram.axis.IntCategory, boost_histogram.axis.StrCategory)
+    ):
+        return None
+
+    labels = [str(label) for label in axis]
+    if isinstance(axis, boost_histogram.axis.IntCategory):
+        # Check labels are valid integers (this may be redundant)
+        for label in labels:
+            try:
+                int(label)
+            except ValueError:
+                raise ValueError(
+                    f"IntCategory labels must be valid integers. Found {label!r} on axis {axis!r}"
+                ) from None
+
+    labels = to_THashList([to_TObjString(label) for label in labels])
+    # we need to set the TObject.fUniqueID to the index of the bin as done by TAxis::SetBinLabel
+    for i, label in enumerate(labels):
+        label._bases[0]._members["@fUniqueID"] = i + 1
+
+    return labels
+
+
 def _root_stats_1d(entries, edges):
     centers = (edges[:-1] + edges[1:]) / 2.0
 
@@ -736,12 +824,14 @@ def to_TObjString(string):
     This function is for developers to create TObjString objects that can be
     written to ROOT files, to implement conversion routines.
     """
+    tobject = uproot.models.TObject.Model_TObject.empty()
+
     tobjstring = uproot.models.TObjString.Model_TObjString(str(string))
     tobjstring._deeply_writable = True
     tobjstring._cursor = None
     tobjstring._parent = None
     tobjstring._members = {}
-    tobjstring._bases = (uproot.models.TObject.Model_TObject(),)
+    tobjstring._bases = (tobject,)
     tobjstring._num_bytes = len(string) + (1 if len(string) < 255 else 5) + 16
     tobjstring._instance_version = 1
     return tobjstring
@@ -762,8 +852,6 @@ def to_TList(data, name=""):
         )
 
     tobject = uproot.models.TObject.Model_TObject.empty()
-    tobject._members["@fUniqueID"] = 0
-    tobject._members["@fBits"] = 0
 
     tlist = uproot.models.TList.Model_TList.empty()
     tlist._bases.append(tobject)
@@ -778,6 +866,30 @@ def to_TList(data, name=""):
     return tlist
 
 
+def to_THashList(data, name=""):
+    """
+    Args:
+        data (:doc:`uproot.model.Model`): Python iterable to convert into a THashList.
+        name (str): Name of the list (usually empty: ``""``).
+
+    This function is for developers to create THashList objects that can be
+    written to ROOT files, to implement conversion routines.
+    """
+
+    if not all(isinstance(x, uproot.model.Model) for x in data):
+        raise TypeError(
+            "list to convert to THashList must only contain ROOT objects (uproot.Model)"
+        )
+
+    tlist = to_TList(data, name)
+
+    thashlist = uproot.models.THashList.Model_THashList.empty()
+
+    thashlist._bases.append(tlist)
+
+    return thashlist
+
+
 def to_TArray(data):
     """
     Args:
@@ -805,9 +917,7 @@ def to_TArray(data):
         cls = uproot.models.TArray.Model_TArrayD
     else:
         raise ValueError(
-            "data to convert to TArray must have signed integer or floating-point type, not {}".format(
-                repr(data.dtype)
-            )
+            f"data to convert to TArray must have signed integer or floating-point type, not {data.dtype!r}"
         )
 
     tarray = cls.empty()
@@ -875,8 +985,6 @@ def to_TAxis(
     written to ROOT files, to implement conversion routines.
     """
     tobject = uproot.models.TObject.Model_TObject.empty()
-    tobject._members["@fUniqueID"] = 0
-    tobject._members["@fBits"] = 0
 
     tnamed = uproot.models.TNamed.Model_TNamed.empty()
     tnamed._deeply_writable = True
@@ -1019,8 +1127,6 @@ def to_TH1x(
     TH1C, TH1D, TH1F, TH1I, or TH1S depends on the dtype of the ``data`` array.
     """
     tobject = uproot.models.TObject.Model_TObject.empty()
-    tobject._members["@fUniqueID"] = 0
-    tobject._members["@fBits"] = 0
 
     tnamed = uproot.models.TNamed.Model_TNamed.empty()
     tnamed._deeply_writable = True
