@@ -1,4 +1,4 @@
-# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE
+# BSD 3-Clause License; see https://github.com/scikit-hep/uproot5/blob/main/LICENSE
 
 """
 This module defines high-level functions and objects for file-writing.
@@ -16,14 +16,16 @@ There is no feature parity between writable and readable versions of each of the
 types. Writing and reading are considered separate projects with different capabilities.
 """
 
+from __future__ import annotations
 
 import datetime
 import itertools
-import os
 import queue
 import sys
 import uuid
 from collections.abc import Mapping, MutableMapping
+from pathlib import Path
+from typing import IO
 
 import uproot._util
 import uproot.compression
@@ -36,18 +38,15 @@ import uproot.writing._cascade
 from uproot._util import no_filter, no_rename
 
 
-def create(file_path, **options):
+def create(file_path: str | Path | IO, **options):
     """
     Args:
         file_path (str, ``pathlib.Path`` or file-like object): The filesystem path of the
             file to open or an open file.
-        compression (:doc:`uproot.compression.Compression` or None): Compression algorithm
-            and level for new objects added to the file. Can be updated after creating
-            the :doc:`uproot.writing.writable.WritableFile`. Default is ``uproot.ZLIB(1)``.
         options: See below.
 
     Opens a local file for writing. Like ROOT's ``"CREATE"`` option, this function
-    raises an error (``OSError``) if a file already exists at ``file_path``.
+    raises an error (``FileExistsError``) if a file already exists at ``file_path``.
 
     Returns a :doc:`uproot.writing.writable.WritableDirectory`.
 
@@ -56,27 +55,33 @@ def create(file_path, **options):
     * initial_directory_bytes (int; 256)
     * initial_streamers_bytes (int; 1024)
     * uuid_function (callable; ``uuid.uuid1``)
+    * compression (:doc:`uproot.compression.Compression` or None): Compression algorithm
+    and level for new objects added to the file. Can be updated after creating
+    the :doc:`uproot.writing.writable.WritableFile`. Default is ``uproot.ZLIB(1)``.
 
     See :doc:`uproot.writing.writable.WritableFile` for details on these options.
+
+    Additional options are passed to as ``storage_options`` to the fsspec filesystem
     """
     file_path = uproot._util.regularize_path(file_path)
-    if uproot._util.isstr(file_path):
-        if os.path.exists(file_path):
-            raise OSError(
-                "path exists and refusing to overwrite (use 'uproot.recreate' to "
-                "overwrite)\n\nfor path {}".format(file_path)
-            )
+    storage_options = {
+        key: value for key, value in options.items() if key not in create.defaults
+    }
+    if isinstance(file_path, str) and uproot.sink.file.FileSink._file_exists(
+        file_path, **storage_options
+    ):
+        raise FileExistsError(
+            "path exists and refusing to overwrite (use 'uproot.recreate' to "
+            f"overwrite)\n\nfor path {file_path}"
+        )
     return recreate(file_path, **options)
 
 
-def recreate(file_path, **options):
+def recreate(file_path: str | Path | IO, **options):
     """
     Args:
         file_path (str, ``pathlib.Path`` or file-like object): The filesystem path of the
             file to open or an open file.
-        compression (:doc:`uproot.compression.Compression` or None): Compression algorithm
-            and level for new objects added to the file. Can be updated after creating
-            the :doc:`uproot.writing.writable.WritableFile`. Default is ``uproot.ZLIB(1)``.
         options: See below.
 
     Opens a local file for writing. Like ROOT's ``"RECREATE"`` option, this function
@@ -89,17 +94,20 @@ def recreate(file_path, **options):
     * initial_directory_bytes (int; 256)
     * initial_streamers_bytes (int; 1024)
     * uuid_function (callable; ``uuid.uuid1``)
+    * compression (:doc:`uproot.compression.Compression` or None): Compression algorithm
+    and level for new objects added to the file. Can be updated after creating
+    the :doc:`uproot.writing.writable.WritableFile`. Default is ``uproot.ZLIB(1)``.
 
     See :doc:`uproot.writing.writable.WritableFile` for details on these options.
-    """
-    file_path = uproot._util.regularize_path(file_path)
-    if uproot._util.isstr(file_path):
-        # Truncate file
-        open(file_path, "w").close()
-        sink = uproot.sink.file.FileSink(file_path)
-    else:
-        sink = uproot.sink.file.FileSink.from_object(file_path)
 
+    Additional options are passed to as ``storage_options`` to the fsspec filesystem.
+    """
+
+    file_path = uproot._util.regularize_path(file_path)
+    storage_options = {
+        key: value for key, value in options.items() if key not in recreate.defaults
+    }
+    sink = uproot.sink.file.FileSink(file_path, **storage_options)
     compression = options.pop("compression", create.defaults["compression"])
 
     initial_directory_bytes = options.pop(
@@ -109,7 +117,7 @@ def recreate(file_path, **options):
         "initial_streamers_bytes", create.defaults["initial_streamers_bytes"]
     )
     uuid_function = options.pop("uuid_function", create.defaults["uuid_function"])
-    if len(options) != 0:
+    if options:
         raise TypeError(
             "unrecognized options for uproot.create or uproot.recreate: "
             + ", ".join(repr(x) for x in options)
@@ -127,7 +135,7 @@ def recreate(file_path, **options):
     ).root_directory
 
 
-def update(file_path, **options):
+def update(file_path: str | Path | IO, **options):
     """
     Args:
         file_path (str, ``pathlib.Path`` or file-like object): The filesystem path of the
@@ -146,18 +154,21 @@ def update(file_path, **options):
     * uuid_function (callable; ``uuid.uuid1``)
 
     See :doc:`uproot.writing.writable.WritableFile` for details on these options.
+
+    Additional options are passed to as ``storage_options`` to the fsspec filesystem
     """
+
     file_path = uproot._util.regularize_path(file_path)
-    if uproot._util.isstr(file_path):
-        sink = uproot.sink.file.FileSink(file_path)
-    else:
-        sink = uproot.sink.file.FileSink.from_object(file_path)
+    storage_options = {
+        key: value for key, value in options.items() if key not in update.defaults
+    }
+    sink = uproot.sink.file.FileSink(file_path, **storage_options)
 
     initial_directory_bytes = options.pop(
         "initial_directory_bytes", create.defaults["initial_directory_bytes"]
     )
     uuid_function = options.pop("uuid_function", create.defaults["uuid_function"])
-    if len(options) != 0:
+    if options:
         raise TypeError(
             "unrecognized options for uproot.update: "
             + ", ".join(repr(x) for x in options)
@@ -213,12 +224,13 @@ class WritableFile(uproot.reading.CommonFileMethods):
         self._fUUID = self._cascading.fileheader.uuid.bytes
 
         self._trees = {}
+        self._ntuples = {}
 
     def __repr__(self):
         return f"<WritableFile {self.file_path!r} at 0x{id(self):012x}>"
 
     @property
-    def sink(self):
+    def sink(self) -> uproot.sink.file.FileSink:
         """
         Returns a :doc:`uproot.sink.file.FileSink`, the physical layer for writing
         (and sometimes reading) data.
@@ -226,7 +238,7 @@ class WritableFile(uproot.reading.CommonFileMethods):
         return self._sink
 
     @property
-    def initial_directory_bytes(self):
+    def initial_directory_bytes(self) -> int:
         """
         Number of bytes to allocate for new directories, so that TKeys can be added
         to them without immediately needing to rewrite the block.
@@ -261,14 +273,14 @@ class WritableFile(uproot.reading.CommonFileMethods):
         }
 
     @property
-    def is_64bit(self):
+    def is_64bit(self) -> bool:
         """
         True if the file has 8-byte pointers in its header; False if the pointers are 4-byte.
         """
         return self._cascading.fileheader.big
 
     @property
-    def compression(self):
+    def compression(self) -> uproot.compression.Compression | None:
         """
         Compression algorithm and level (:doc:`uproot.compression.Compression` or None)
         for new objects added to the file.
@@ -300,7 +312,7 @@ class WritableFile(uproot.reading.CommonFileMethods):
         return self._cascading.fileheader.free_location
 
     @property
-    def fNbytesFree(self):
+    def fNbytesFree(self) -> int:
         """
         The number of bytes in the ``TFree`` data, for managing empty spaces
         in a ROOT file (filesystem-like fragmentation).
@@ -308,7 +320,7 @@ class WritableFile(uproot.reading.CommonFileMethods):
         return self._cascading.fileheader.free_num_bytes
 
     @property
-    def nfree(self):
+    def nfree(self) -> int:
         """
         The number of objects in the ``TFree`` data, for managing empty spaces
         in a ROOT file (filesystem-like fragmentation).
@@ -316,7 +328,7 @@ class WritableFile(uproot.reading.CommonFileMethods):
         return self._cascading.fileheader.free_num_slices + 1
 
     @property
-    def fUnits(self):
+    def fUnits(self) -> int:
         """
         Number of bytes in the serialization of file seek points, which can either
         be 4 or 8.
@@ -348,7 +360,7 @@ class WritableFile(uproot.reading.CommonFileMethods):
         return self._cascading.fileheader.info_location
 
     @property
-    def fNbytesInfo(self):
+    def fNbytesInfo(self) -> int:
         """
         The number of bytes in the ``TStreamerInfo`` data, where
         TStreamerInfo records are located.
@@ -378,7 +390,7 @@ class WritableFile(uproot.reading.CommonFileMethods):
         self._cascading.streamers.update_streamers(self.sink, streamers)
 
     @property
-    def file_path(self):
+    def file_path(self) -> str | None:
         """
         Filesystem path of the open file, or None if using a file-like object.
         """
@@ -396,7 +408,7 @@ class WritableFile(uproot.reading.CommonFileMethods):
         self._sink.close()
 
     @property
-    def closed(self):
+    def closed(self) -> bool:
         """
         True if the file has been closed; False otherwise.
 
@@ -418,12 +430,21 @@ class WritableFile(uproot.reading.CommonFileMethods):
     def _new_tree(self, tree):
         self._trees[tree._cascading.key.seek_location] = tree
 
+    def _new_ntuple(self, ntuple):
+        self._ntuples[ntuple._cascading.key.seek_location] = ntuple
+
     def _has_tree(self, loc):
         return loc in self._trees
 
     def _get_tree(self, loc):
         return self._trees[loc]
 
+    def _has_ntuple(self, loc):
+        return loc in self._ntuples
+
+    def _get_ntuple(self, loc):
+        return self._ntuples[loc]
+
     def _move_tree(self, oldloc, newloc):
         tree = self._trees[oldloc]
         del self._trees[oldloc]
@@ -524,7 +545,7 @@ class WritableDirectory(MutableMapping):
         Path of directory names to this subdirectory as a single string, delimited
         by slashes.
         """
-        return "/".join(("",) + self._path + ("",)).replace("//", "/")
+        return "/".join(("", *self._path, "")).replace("//", "/")
 
     @property
     def file_path(self):
@@ -553,7 +574,7 @@ class WritableDirectory(MutableMapping):
         self._file.close()
 
     @property
-    def closed(self):
+    def closed(self) -> bool:
         """
         True if the file has been closed; False otherwise.
 
@@ -595,10 +616,7 @@ class WritableDirectory(MutableMapping):
     def __contains__(self, where):
         if self._cascading.data.haskey(where):
             return True
-        for x in self._cascading.data.dir_names:
-            if where in self._subdir(x):
-                return True
-        return False
+        return any(where in self._subdir(x) for x in self._cascading.data.dir_names)
 
     def __iter__(self):
         return self.iterkeys()
@@ -611,6 +629,7 @@ class WritableDirectory(MutableMapping):
 
     def keys(
         self,
+        *,
         recursive=True,
         cycle=True,
         filter_name=no_filter,
@@ -643,6 +662,7 @@ class WritableDirectory(MutableMapping):
 
     def values(
         self,
+        *,
         recursive=True,
         filter_name=no_filter,
         filter_classname=no_filter,
@@ -673,6 +693,7 @@ class WritableDirectory(MutableMapping):
 
     def items(
         self,
+        *,
         recursive=True,
         cycle=True,
         filter_name=no_filter,
@@ -706,6 +727,7 @@ class WritableDirectory(MutableMapping):
 
     def classnames(
         self,
+        *,
         recursive=True,
         cycle=True,
         filter_name=no_filter,
@@ -738,6 +760,7 @@ class WritableDirectory(MutableMapping):
 
     def iterkeys(
         self,
+        *,
         recursive=True,
         cycle=True,
         filter_name=no_filter,
@@ -784,6 +807,7 @@ class WritableDirectory(MutableMapping):
 
     def itervalues(
         self,
+        *,
         recursive=True,
         filter_name=no_filter,
         filter_classname=no_filter,
@@ -814,6 +838,7 @@ class WritableDirectory(MutableMapping):
 
     def iteritems(
         self,
+        *,
         recursive=True,
         cycle=True,
         filter_name=no_filter,
@@ -844,11 +869,12 @@ class WritableDirectory(MutableMapping):
         ):
             if not cycle:
                 at = keyname.index(";")
-                keyname = keyname[:at]
+                keyname = keyname[:at]  # noqa: PLW2901 (overwriting keyname)
             yield keyname, self[keyname]
 
     def iterclassnames(
         self,
+        *,
         recursive=True,
         cycle=True,
         filter_name=no_filter,
@@ -983,6 +1009,13 @@ class WritableDirectory(MutableMapping):
                 raise TypeError(
                     "WritableDirectory cannot view preexisting TTrees; open the file with uproot.open instead of uproot.recreate or uproot.update"
                 )
+        elif key.classname.string == "ROOT::RNTuple":
+            if self._file._has_ntuple(key.seek_location):
+                return self._file._get_ntuple(key.seek_location)
+            else:
+                raise TypeError(
+                    "WritableDirectory cannot view preexisting RNTuple; open the file with uproot.open instead of uproot.recreate or uproot.update"
+                )
 
         else:
 
@@ -1015,7 +1048,7 @@ class WritableDirectory(MutableMapping):
 
     def _del(self, name, cycle):
         key = self._cascading.data.get_key(name, cycle)
-        start = key.location
+        start = key.seek_location
         stop = start + key.num_bytes + key.compressed_bytes
         self._cascading.freesegments.release(start, stop)
 
@@ -1038,7 +1071,7 @@ class WritableDirectory(MutableMapping):
                     and tree._cascading.directory is not sub._cascading
                 ):
                     self._subdirs[name] = WritableDirectory(
-                        self._path + (name,), self._file, tree._cascading.directory
+                        (*self._path, name), self._file, tree._cascading.directory
                     )
                     break
 
@@ -1120,7 +1153,7 @@ class WritableDirectory(MutableMapping):
                 self._file.sink.flush()
 
                 self._subdirs[name] = WritableDirectory(
-                    self._path + (name,), self._file, subdirectory
+                    (*self._path, name), self._file, subdirectory
                 )
 
             else:
@@ -1147,12 +1180,12 @@ class WritableDirectory(MutableMapping):
                 )
 
                 self._subdirs[name] = WritableDirectory(
-                    self._path + (name,), self._file, subdirectory
+                    (*self._path, name), self._file, subdirectory
                 )
 
         return self._subdirs[name]
 
-    def mkdir(self, name, initial_directory_bytes=None):
+    def mkdir(self, name, *, initial_directory_bytes=None):
         """
         Args:
             name (str): Name of the new subdirectory.
@@ -1189,7 +1222,7 @@ class WritableDirectory(MutableMapping):
             if initial_directory_bytes is None:
                 initial_directory_bytes = self._file.initial_directory_bytes
             directory = WritableDirectory(
-                self._path + (head,),
+                (*self._path, head),
                 self._file,
                 self._cascading.add_directory(
                     self._file.sink,
@@ -1201,10 +1234,8 @@ class WritableDirectory(MutableMapping):
 
         elif key.classname.string not in ("TDirectory", "TDirectoryFile"):
             raise TypeError(
-                """cannot make a directory named {} because a {} already has that name
-in file {} in directory {}""".format(
-                    repr(name), key.classname.string, self.file_path, self.path
-                )
+                f"""cannot make a directory named {name!r} because a {key.classname.string} already has that name
+in file {self.file_path} in directory {self.path}"""
             )
 
         else:
@@ -1221,6 +1252,7 @@ in file {} in directory {}""".format(
         name,
         branch_types,
         title="",
+        *,
         counter_name=lambda counted: "n" + counted,
         field_name=lambda outer, inner: inner if outer == "" else outer + "_" + inner,
         initial_basket_capacity=10,
@@ -1266,7 +1298,7 @@ in file {} in directory {}""".format(
             dirpath, treename = name[:at], name[at + 1 :]
             directory = self.mkdir(dirpath)
 
-        path = directory._path + (treename,)
+        path = (*directory._path, treename)
 
         tree = WritableTree(
             path,
@@ -1312,9 +1344,62 @@ in file {} in directory {}""".format(
 
         return tree
 
+    def mkrntuple(
+        self,
+        name,
+        ak_form_or_data,
+        title="",
+    ):
+        """
+        Args:
+            name (str): Name of the new RNTuple.
+            ak_form_or_data (Awkward RecordForm or RecordArray): Name
+                and type specification for the fields. If a RecordForm is provided,
+                the RNTuple will be empty. If a RecordArray is provided, the RNTuple
+                will be initialized with the input data.
+            title (str): Title for the new RNTuple.
+
+        Creates an empty RNTuple in this directory.
+        """
+        if self._file.sink.closed:
+            raise ValueError("cannot create a RNTuple in a closed file")
+
+        # TODO: Think of a better alternative to this
+        if isinstance(ak_form_or_data, uproot.extras.awkward().Array):
+            ntuple = self.mkrntuple(name, ak_form_or_data.layout.form, title)
+            ntuple.extend(ak_form_or_data)
+            return ntuple
+
+        # The rest assumes that ak_form_or_data is a RecordForm
+
+        try:
+            at = name.rindex("/")
+        except ValueError:
+            treename = name
+            directory = self
+        else:
+            dirpath, treename = name[:at], name[at + 1 :]
+            directory = self.mkdir(dirpath)
+
+        path = (*directory._path, treename)
+
+        ntuple = WritableNTuple(
+            path,
+            directory._file,
+            directory._cascading.add_rntuple(
+                directory._file.sink,
+                treename,
+                title,
+                ak_form_or_data,
+            ),
+        )
+        directory._file._new_ntuple(ntuple)
+        return ntuple
+
     def copy_from(
         self,
         source,
+        *,
         filter_name=no_filter,
         filter_classname=no_filter,
         rename=no_rename,
@@ -1354,10 +1439,8 @@ in file {} in directory {}""".format(
         if len(old_names) == 0:
             if require_matches:
                 raise ValueError(
-                    """no objects found with names matching {}
-in file {} in directory {}""".format(
-                        repr(filter_name), source.file_path, source.path
-                    )
+                    f"""no objects found with names matching {filter_name!r}
+in file {source.file_path} in directory {source.path}"""
                 )
             else:
                 return
@@ -1414,7 +1497,12 @@ in file {} in directory {}""".format(
             )
 
         for name, allocation in new_dirs.items():
-            self.mkdir(name, max(self._file.initial_directory_bytes, allocation))
+            self.mkdir(
+                name,
+                initial_directory_bytes=max(
+                    self._file.initial_directory_bytes, allocation
+                ),
+            )
 
         for _ in range(len(ranges)):
             chunk = notifications.get()
@@ -1463,7 +1551,10 @@ in file {} in directory {}""".format(
             path, name = fullpath[:-1], fullpath[-1]
 
             if len(path) != 0:
-                self.mkdir("/".join(path), self._file.initial_directory_bytes)
+                self.mkdir(
+                    "/".join(path),
+                    initial_directory_bytes=self._file.initial_directory_bytes,
+                )
 
             directory = self
             for item in path:
@@ -1572,15 +1663,15 @@ class WritableTree:
         return self._path
 
     @property
-    def object_path(self):
+    def object_path(self) -> str:
         """
         Path of directory names to this TTree as a single string, delimited by
         slashes.
         """
-        return "/".join(("",) + self._path + ("",)).replace("//", "/")
+        return "/".join(("", *self._path, "")).replace("//", "/")
 
     @property
-    def file_path(self):
+    def file_path(self) -> str | None:
         """
         Filesystem path of the open file, or None if using a file-like object.
         """
@@ -1606,7 +1697,7 @@ class WritableTree:
         self._file.close()
 
     @property
-    def closed(self):
+    def closed(self) -> bool:
         """
         True if the file has been closed; False otherwise.
 
@@ -1669,7 +1760,7 @@ class WritableTree:
         elif (
             isinstance(value, Mapping)
             and all(
-                uproot._util.isstr(k)
+                isinstance(k, str)
                 and (v is None or isinstance(v, uproot.compression.Compression))
                 for k, v in value.items()
             )
@@ -1708,14 +1799,14 @@ class WritableTree:
             )
 
     @property
-    def num_entries(self):
+    def num_entries(self) -> int:
         """
         The number of entries accumulated so far.
         """
         return self._cascading.num_entries
 
     @property
-    def num_baskets(self):
+    def num_baskets(self) -> int:
         """
         The number of TBaskets accumulated so far.
         """
@@ -1746,12 +1837,13 @@ class WritableTree:
 
         .. warning::
 
-            **As a word of warning,** be sure that each call to :ref:`uproot.writing.writable.WritableTree.extend` includes at least 100 kB per branch/array. (NumPy and Awkward Arrays have an `nbytes <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.nbytes.html>`__ property; you want at least ``100000`` per array.) If you ask Uproot to write very small TBaskets, it will spend more time working on TBasket overhead than actually writing data. The absolute worst case is one-entry-per-:ref:`uproot.writing.writable.WritableTree.extend`. See `#428 (comment) <https://github.com/scikit-hep/uproot4/pull/428#issuecomment-908703486>`__.
+            **As a word of warning,** be sure that each call to :ref:`uproot.writing.writable.WritableTree.extend` includes at least 100 kB per branch/array. (NumPy and Awkward Arrays have an `nbytes <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.nbytes.html>`__ property; you want at least ``100000`` per array.) If you ask Uproot to write very small TBaskets, it will spend more time working on TBasket overhead than actually writing data. The absolute worst case is one-entry-per-:ref:`uproot.writing.writable.WritableTree.extend`. See `#428 (comment) <https://github.com/scikit-hep/uproot5/pull/428#issuecomment-908703486>`__.
         """
         self._cascading.extend(self._file, self._file.sink, data)
 
     def show(
         self,
+        *,
         filter_name=no_filter,
         filter_typename=no_filter,
         filter_branch=no_filter,
@@ -1850,3 +1942,226 @@ class WritableBranch:
             raise TypeError(
                 "compression must be None or a uproot.compression.Compression object, like uproot.ZLIB(4) or uproot.ZSTD(0)"
             )
+
+
+class WritableNTuple:
+    """
+    Args:
+        path (tuple of str): Path of directory names to this RNTuple.
+        file (:doc:`uproot.writing.writable.WritableFile`): Handle to the file in
+            which this RNTuple can be found.
+        cascading (:doc:`uproot.writing._cascadentuple.NTuple`): The low-level
+            directory object.
+
+    Represents a writable ``RNTuple`` from a ROOT file.
+
+    Assigning TTree-like data to a directory creates the TTree object with all of
+    its metadata and fills it with the contents of the arrays in one step. To separate
+    the process of creating the TTree metadata from filling the first TBasket, use the
+    :doc:`uproot.writing.writable.WritableDirectory.mktree` method:
+
+    .. code-block:: python
+
+        my_directory.mkrntuple("tuple6", {"branch1": numpy_dtype, "branch2": awkward_type})
+
+    The ``numpy_dtype`` is any data that NumPy recognizes as a ``np.dtype``, and the
+    ``awkward_type`` is an `ak.types.Type <https://awkward-array.readthedocs.io/en/latest/ak.types.Type.html>`__ from
+    `ak.type <https://awkward-array.readthedocs.io/en/latest/_auto/ak.type.html>`__ or
+    a string in that form, such as ``"var * float64"`` for variable-length doubles.
+
+    RNTuple can be extended using :ref:`uproot.writing.writable.WritableNTuple.extend`
+    method:
+
+    .. code-block:: python
+
+        my_directory["tuple6"].extend({"branch1": another_numpy_array,
+                                      "branch2": another_awkward_array})
+
+    Be sure to make these extensions as large as is feasible within memory constraints,
+    because a ROOT file full of small clusters is bloated (larger than it needs to be)
+    and slow to read (especially for Uproot, but also for ROOT).
+
+    For instance, if you want to write a million events and have enough memory
+    available to do that 100 thousand events at a time (total of 10 TBaskets),
+    then do so. Filling the RNTuple a hundred events at a time (total of 10000 TBaskets)
+    would be considerably slower for writing and reading, and the file would be much
+    larger than it could otherwise be, even with compression.
+    """
+
+    def __init__(self, path, file, cascading):
+        self._path = path
+        self._file = file
+        self._cascading = cascading
+
+    def __repr__(self):
+        return "<WritableNTuple {} at 0x{:012x}>".format(
+            repr("/" + "/".join(self._path)), id(self)
+        )
+
+    @property
+    def path(self):
+        """
+        Path of directory names to this RNTuple as a tuple of strings.
+        """
+        return self._path
+
+    @property
+    def object_path(self):
+        """
+        Path of directory names to this RNTuple as a single string, delimited by
+        slashes.
+        """
+        return "/".join(("", *self._path, "")).replace("//", "/")
+
+    @property
+    def file_path(self) -> str | None:
+        """
+        Filesystem path of the open file, or None if using a file-like object.
+        """
+        return self._file.file_path
+
+    @property
+    def file(self):
+        """
+        Handle to the :doc:`uproot.writing.writable.WritableDirectory` in which
+        this directory can be found.
+        """
+        return self._file
+
+    def close(self):
+        """
+        Explicitly close the file.
+
+        (Files can also be closed with the Python ``with`` statement, as context
+        managers.)
+
+        After closing, objects cannot be read from or written to the file.
+        """
+        self._file.close()
+
+    @property
+    def closed(self) -> bool:
+        """
+        True if the file has been closed; False otherwise.
+
+        The file may have been closed explicitly with
+        :ref:`uproot.writing.writable.WritableFile.close` or implicitly in the Python
+        ``with`` statement, as a context manager.
+
+        After closing, objects cannot be read from or written to the file.
+        """
+        return self._file.closed
+
+    def __enter__(self):
+        self._file.sink.__enter__()
+        return self
+
+    def __exit__(self, exception_type, exception_value, traceback):
+        self._file.sink.__exit__(exception_type, exception_value, traceback)
+
+    @property
+    def compression(self):
+        """
+        Compression algorithm and level (:doc:`uproot.compression.Compression` or None)
+        for new blobs added to the RNTuple.
+
+        This property can be changed and doesn't have to be the same as the compression
+        of the file, which allows you to write different objects with different
+        compression settings.
+
+        The following are equivalent:
+
+        .. code-block:: python
+
+            my_directory["tree"]["branch1"].compression = uproot.ZLIB(1)
+            my_directory["tree"]["branch2"].compression = uproot.LZMA(9)
+
+        and
+
+        .. code-block:: python
+
+            my_directory["tree"].compression = {"branch1": uproot.ZLIB(1),
+                                                "branch2": uproot.LZMA(9)}
+        """
+        out = {}
+        last = None
+        for datum in self._cascading._branch_data:
+            if datum["kind"] != "record":
+                last = out[datum["fName"]] = datum["compression"]
+        if all(x == last for x in out.values()):
+            return last
+        else:
+            return out
+
+    @compression.setter
+    def compression(self, value):
+        if value is None or isinstance(value, uproot.compression.Compression):
+            for datum in self._cascading._branch_data:
+                if datum["kind"] != "record":
+                    datum["compression"] = value
+
+        elif (
+            isinstance(value, Mapping)
+            and all(
+                isinstance(k, str)
+                and (v is None or isinstance(v, uproot.compression.Compression))
+                for k, v in value.items()
+            )
+            and all(
+                datum["fName"] in value
+                for datum in self._cascading._branch_data
+                if datum["kind"] != "record"
+            )
+            and len(value)
+            == len(
+                [
+                    datum
+                    for datum in self._cascading._branch_data
+                    if datum["kind"] != "record"
+                ]
+            )
+        ):
+            for datum in self._cascading._branch_data:
+                if datum["kind"] != "record":
+                    datum["compression"] = value[datum["fName"]]
+
+        else:
+            raise TypeError(
+                "compression must be None, a uproot.compression.Compression object, like uproot.ZLIB(4) or uproot.ZSTD(0), or a mapping of branch names to such objects"
+            )
+
+    @property
+    def num_entries(self) -> int:
+        """
+        The number of entries accumulated so far.
+        """
+        return self._cascading.num_entries
+
+    def extend(self, data):
+        """
+        Args:
+            data (dict of str \u2192 arrays): More array data to add to the RNTuple.
+
+        This method adds data to an existing RNTuple, whether it was created through
+        assignment or :doc:`uproot.writing.writable.WritableDirectory.mkrntuple`.
+
+        The arrays must be a dict, but the values of the dict can be any of the
+        array/DataFrame types described in :doc:`uproot.writing.writable.WritableTree`.
+        However, these types must be compatible with the established TBranch
+        types, the dict must contain a key for every TBranch, and the arrays must have
+        the same lengths (in their first dimension).
+
+        For example,
+
+        .. code-block:: python
+
+            my_directory.mkrntuple("ntuple6", {"branch1": numpy_dtype, "branch2": awkward_type})
+
+            my_directory["ntuple6"].extend({"branch1": another_numpy_array,
+                                          "branch2": another_awkward_array})
+
+        .. warning::
+
+            **As a word of warning,** be sure that each call to :ref:`uproot.writing.writable.WritableNTuple.extend` includes at least 100 kB per branch/array. (NumPy and Awkward Arrays have an `nbytes <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.nbytes.html>`__ property; you want at least ``100000`` per array.) If you ask Uproot to write very small TBaskets, it will spend more time working on TBasket overhead than actually writing data. The absolute worst case is one-entry-per-:ref:`uproot.writing.writable.WritableTree.extend`. See `#428 (comment) <https://github.com/scikit-hep/uproot5/pull/428#issuecomment-908703486>`__.
+        """
+        self._cascading.extend(self._file, self._file.sink, data)
